-- Code to manage the statrings configuration (Note tihs is logically
-- indepdendent of the OPTIONS code which manages the gui to control the
-- config).

-- Basic data items

-- UNIT_MAP - Each unit there's a ring for, with its group id
local UNIT_MAP = {
   player    = "player",
   target    = "target",
   pet       = "pet",
   party1    = "party",
   party2    = "party",
   party3    = "party",
   party4    = "party",
   partypet1 = "partypet",
   partypet2 = "partypet",
   partypet3 = "partypet",
   partypet4 = "partypet",
};

-- DEFAULT_SETTINGS - Used to initialize missing values in the various
-- configurations.
local DEFAULT_SETTINGS = {
   reactions = {
      alphaFull = 0.0,
      alphaNormal = 0.5,
      alphaAlert = 0.7,
      healthAlert = 0.5,
      manaUnfade = 1.0,
   }
}

-- Default options
local INITIAL_OPTIONS =  {
   locX = 0.0,
   locY = -0.25,
   scale = 0.5,

   reactions = {
      player = {
	 alphaFull = 0.0,
	 alphaNormal = 0.5,
	 alphaAlert = 0.7,
	 healthAlert = 0.5,
	 manaUnfade = 1.0,
      },
      partypet = {
	 copyOf = "pet"
      }
   }
};

STATRINGS_OPTIONS = INITIAL_OPTIONS;

local CONFIG = {};
StatRingsConfig = CONFIG;

local configInitialized = false;

---------------------------------------------------------------------------


-- Deep copy a table structure, handling self-referential tables properly
-- (Does NOT copy table KEYS, just VALUES)
-- 
-- returns copy
local function deepCopy(data, ret, cache)
   local type = type;
   if (not ret) then
     ret = {};
   end
   for k,v in pairs(data) do
      if (type(v) == "table") then
         if (not cache) then
           cache = { [data] = ret };
         end
	 if (cache[v]) then
	    ret[k] = cache[v];
	 else
	    local subTable = {};
	    ret[k] = subTable;
	    cache[v] = subTable;
	    subTable, cache = deepCopy(v, subTable, cache);
	 end
      else
	 ret[k] = v;
      end
   end
   return ret;
end

---------------------------------------------------------------------------

-- Called when the addon is loaded
function CONFIG:InitConfig() 
   if (INITIAL_OPTIONS ~= STATRINGS_OPTIONS) then
      -- We've been reloaded!!
      if (not STATRINGS_OPTIONS.reactions) then
	 STATRINGS_OPTIONS.reactions = INITIAL_OPTIONS.reactions;
      end
   end

   StatRingsConfig_BlessHierarchialConfig(STATRINGS_OPTIONS.reactions);

   local workData  = STATRINGS_OPTIONS.reactions;
   local defaultData = DEFAULT_SETTINGS.reactions;
   for _,groupData in pairs(workData) do
      if (not groupData.copyOf) then
	 for k,v in pairs(defaultData) do
	    if (not groupData[k]) then
	       groupData[k] = v;
	    end
	 end
      end
   end
   
   configInitialized = true;
   self.workingOptions = {};


   DEFAULT_CHAT_FRAME:AddMessage("StatRings: Configuration options ready...");


   self:ApplyCurrentSettings("reactions");
   StatRingsCluster:ApplyLocationOptions(STATRINGS_OPTIONS);
end

function CONFIG:GetTypeConfig(configType)
   local config = self.workingOptions[configType];
   if (not config) then
      config = STATRINGS_OPTIONS[configType];
   end
   return config;
end

function CONFIG:SetSettings(configType, configGroup, settingTable)
   local config = self.workingOptions[configType];
   if (not config) then
      config = STATRINGS_OPTIONS[configType];
      if (not config) then
	 return nil;
      end
      config = deepCopy(config);
      StatRingsConfig_BlessHierarchialConfig(config);
      self.workingOptions[configType] = config;
   end
   
   if (not config:ApplyConfig(configGroup, settingTable, "player", true)) then
      return;
   end
   
   self:ApplyCurrentSettings(configType);
   return true;
end

function CONFIG:GetSettings(configType, configGroup, resultTable)
   if (resultTable) then
      -- Clear table in case we return nil.
      for k in pairs(resultTable) do
	 resultTable[k] = nil;
      end
   end
   if (not configInitialized) then
      return nil;
   end
   local config = self:GetTypeConfig(configType);
   if (not config) then
      return nil;
   end
   if (not resultTable) then
      resultTable = {};
   end
   return config:GetConfig(configGroup, "player", resultTable);
end

function CONFIG:GetEffectiveSettings(configType, configGroup, resultTable)
   if (resultTable) then
      -- Clear table in case we return nil.
      for k in pairs(resultTable) do
	 resultTable[k] = nil;
      end
   end
   if (not configInitialized) then
      return nil;
   end
   local config = self:GetTypeConfig(configType);
   if (not config) then
      return nil;
   end
   if (not resultTable) then
      resultTable = {};
   end
   return config:GetEffectiveConfig(configGroup, "player", resultTable);
end

local EMPTY_TABLE = {};

-- CONFIG_GROUPS - Map of group id to boolean, used as return value
--                 in a few places
local CONFIG_GROUPS = {};
for k,v in UNIT_MAP do
   CONFIG_GROUPS[v] = false;
end


function CONFIG:GetEligibleCopies(configType, configGroup)
   if (not configInitialized) then
      return EMPTY_TABLE;
   end
   local config = self:GetTypeConfig(configType);
   if (not config) then
      return EMPTY_TABLE;
   end
   return config:CheckCopyEligibility(configGroup, CONFIG_GROUPS, "player");
end

function CONFIG:GetDependentGroups(configType, configGroup)
   if (not configInitialized) then
      return EMPTY_TABLE;
   end
   local config = self:GetTypeConfig(configType);
   if (not config) then
      return EMPTY_TABLE;
   end
   return config:FindDependentGroups(configGroup, CONFIG_GROUPS, "player");
end

local activePreviewMap = {};
local isPreviewActive = nil;

function CONFIG:InitVisualPreview(configType, configGroup,
				  enabledVisual, disabledVisual)
   if (not configInitialized) then
      return;
   end

   local config = self:GetTypeConfig(configType);
   if (config 
       and config[configGroup] and (not config[configGroup].copyOf)) then
      -- This is its own settings
      local groupList = self:GetDependentGroups(configType, configGroup);
      for k,v in pairs(UNIT_MAP) do
	 if (groupList[v]) then
	    activePreviewMap[k] = true;
	 else
	    activePreviewMap[k] = nil;
	 end
      end
   else
      for k,v in pairs(UNIT_MAP) do
	 if (v == configGroup) then
	    activePreviewMap[k] = true;
	 else
	    activePreviewMap[k] = nil;
	 end
      end
   end
   
   isPreviewActive = true;
   StatRingsCluster:PreviewInfoModel(activePreviewMap, 
				     enabledVisual, disabledVisual);
end

function CONFIG:UpdateVisualPreview(enabledVisual)
   if (isPreviewActive) then
      StatRingsCluster:NotifyInfoUpdate(enabledVisual);
   end
end

function CONFIG:ResetVisualPreview()
   if (isPreviewActive) then
      StatRingsCluster:RestoreInfoModel();
      isPreviewActive = nil;
   end
end

local tmpSettings = {};
local groupCache = {};

function CONFIG:DiscardWorkingSettings()
   local workOpts = self.workingOptions;
   for configType in pairs(workOpts) do
      workOpts[configType] = nil;
      self:ApplyCurrentSettings(configType);
   end
end

function CONFIG:ApplyWorkingSettings()
   local workOpts = self.workingOptions;
   for configType,workData in pairs(workOpts) do
      STATRINGS_OPTIONS[configType] = workData;
      workOpts[configType] = nil;

      for _,groupData in pairs(workData) do
	 -- Purge unnecessary data
	 if (groupData.copyOf) then
	    for k in pairs(groupData) do
	       if (k ~= "copyOf") then
		  groupData[k] = nil;
	       end
	    end
	 end
      end
      self:ApplyCurrentSettings(configType);
   end
end

function CONFIG:ApplyCurrentSettings(configType)
   for k,v in pairs(groupCache) do
      v.bad = true;
   end
   for unit,group in UNIT_MAP do
      local settings = groupCache[group];
      local bad;
      if (not settings) then
	 bad = true;
	 settings = {};
	 groupCache[group] = settings;
      else
	 bad = settings.bad;
      end
      if (bad) then
	 self:GetEffectiveSettings(configType, group, settings);
	 if (settings.bad) then
	    for k in pairs(settings) do
	       settings[k] = nil;
	    end
	 end
      end
      
      tmpSettings[unit] = settings;
   end

   StatRingsCluster:ApplyUnitConfigMap(tmpSettings, true);
end
---------------------------------------------------------------------------

